#!/usr/bin/python

from __future__ import print_function

__docformat__ = 'restructedtext en'

import six.moves.cPickle as pickle
import gzip
import os
import sys
import timeit
import numpy
import cPickle
from theano import tensor
import theano
import theano.tensor as T
from theano.tensor.nnet import conv2d
from theano.tensor.nnet.conv3d2d import conv3d
from theano.tensor.signal import pool

class LogisticRegression(object):
    """Multi-class Logistic Regression Class

    The logistic regression is fully described by a weight matrix :math:`W`
    and bias vector :math:`b`. Classification is done by projecting data
    points onto a set of hyperplanes, the distance to which is used to
    determine a class membership probability.
    """

    def __init__(self, input, n_in, n_out, parameters=str("None")):
        """ Initialize the parameters of the logistic regression

        :type input: theano.tensor.TensorType
        :param input: symbolic variable that describes the input of the
                      architecture (one minibatch)

        :type n_in: int
        :param n_in: number of input units, the dimension of the space in
                     which the datapoints lie

        :type n_out: int
        :param n_out: number of output units, the dimension of the space in
                      which the labels lie

        """
        # start-snippet-1

        if str(parameters)==str("None"):
            # initialize with 0 the weights W as a matrix of shape (n_in, n_out)
            self.W = theano.shared(
                value=numpy.zeros(
                    (n_in, n_out),
                    dtype=theano.config.floatX
                ),
                name='W',
                borrow=True
            )
            # initialize the biases b as a vector of n_out 0s
            self.b = theano.shared(
                value=numpy.zeros(
                    (n_out,),
                    dtype=theano.config.floatX
                ),
                name='b',
                borrow=True
            )

        else:
            self.W, self.b = cPickle.load(open(str(parameters)+'.pkl')).params
            print("Parameters from "+str(parameters)+" extracted.")

        # symbolic expression for computing the matrix of class-membership
        # probabilities
        # Where:
        # W is a matrix where column-k represent the separation hyperplane for
        # class-k
        # x is a matrix where row-j  represents input training sample-j
        # b is a vector where element-k represent the free parameter of
        # hyperplane-k
        self.p_y_given_x = T.nnet.softmax(T.dot(input, self.W) + self.b)

            # symbolic description of how to compute prediction as class whose
        # probability is maximal
        self.y_pred = T.argmax(self.p_y_given_x, axis=1)
        # end-snippet-1

        # parameters of the model
        self.params = [self.W, self.b]

        # keep track of model input
        self.input = input

    def negative_log_likelihood(self, y):
        """Return the mean of the negative log-likelihood of the prediction
        of this model under a given target distribution.

        .. math::

            \frac{1}{|\mathcal{D}|} \mathcal{L} (\theta=\{W,b\}, \mathcal{D}) =
            \frac{1}{|\mathcal{D}|} \sum_{i=0}^{|\mathcal{D}|}
                \log(P(Y=y^{(i)}|x^{(i)}, W,b)) \\
            \ell (\theta=\{W,b\}, \mathcal{D})

        :type y: theano.tensor.TensorType
        :param y: corresponds to a vector that gives for each example the
                  correct label

        Note: we use the mean instead of the sum so that
              the learning rate is less dependent on the batch size
        """
        # start-snippet-2
        # y.shape[0] is (symbolically) the number of rows in y, i.e.,
        # number of examples (call it n) in the minibatch
        # T.arange(y.shape[0]) is a symbolic vector which will contain
        # [0,1,2,... n-1] T.log(self.p_y_given_x) is a matrix of
        # Log-Probabilities (call it LP) with one row per example and
        # one column per class LP[T.arange(y.shape[0]),y] is a vector
        # v containing [LP[0,y[0]], LP[1,y[1]], LP[2,y[2]], ...,
        # LP[n-1,y[n-1]]] and T.mean(LP[T.arange(y.shape[0]),y]) is
        # the mean (across minibatch examples) of the elements in v,
        # i.e., the mean log-likelihood across the minibatch.
        return -T.mean(T.log(self.p_y_given_x)[T.arange(y.shape[0]), y])
        # end-snippet-2

    def errors(self, y):
        """Return a float representing the number of errors in the minibatch
        over the total number of examples of the minibatch ; zero one
        loss over the size of the minibatch

        :type y: theano.tensor.TensorType
        :param y: corresponds to a vector that gives for each example the
                  correct label
        """

        # check if y has same dimension of y_pred
        if y.ndim != self.y_pred.ndim:
            raise TypeError(
                'y should have the same shape as self.y_pred',
                ('y', y.type, 'y_pred', self.y_pred.type)
            )
        # check if y is of the correct datatype
        if y.dtype.startswith('int'):
            # the T.neq operator returns a vector of 0s and 1s, where 1
            # represents a mistake in prediction
            return T.mean(T.neq(self.y_pred, y))
	# T.neq(a,b): =1 if a!=b, and =0 if a==b.
        else:
            raise NotImplementedError()

class MSERegression(object):
    def __init__(self, input, n_in, n_out, parameters=str("None")):
        if str(parameters)==str("None"):
            self.W = theano.shared(
                value=numpy.ones(
                    (n_in, n_out),
                    dtype=theano.config.floatX
                ),
                name='W',
                borrow=True
            )  
            # initialize the biases b as a vector of n_out 0s
            self.b = theano.shared(
                value=numpy.ones(
                    (n_out,),
                    dtype=theano.config.floatX
                ),
                name='b',
                borrow=True
            )
        else:
            self.W, self.b = cPickle.load(open(str(parameters)+'.pkl')).params
            print("Parameters from "+str(parameters)+" extracted.")

        self.params = [self.W, self.b]
#        self.p_y_given_x = T.nnet.sigmoid(T.dot(input, self.W) + self.b).reshape(input.shape[0],n_out)
#        self.y_pred = T.nnet.sigmoid(T.dot(input, self.W) + self.b).dimshuffle(1,0)[0]
        self.p_y_given_x = (T.dot(input, self.W) + self.b).reshape(input.shape[0],n_out)
        self.y_pred = (T.dot(input, self.W) + self.b).dimshuffle(1,0)[0]
        self.input = input

    def mean_squared_error(self, y):
        return T.mean(T.pow((self.p_y_given_x - y),2))


class HiddenLayer(object):
    def __init__(self, rng, input, n_in, n_out, W=None, b=None,
                 activation=T.tanh, parameters=str("None")):
        """
	Typical hidden layer of a MLP: units are fully-connected and have
        sigmoidal activation function. Weight matrix W is of shape (n_in,n_out)
        and the bias vector b is of shape (n_out,).

        NOTE : The nonlinearity used here is tanh

        Hidden unit activation is given by: tanh(dot(input,W) + b)

        :type rng: numpy.random.RandomState
        :param rng: a random number generator used to initialize weights

        :type input: theano.tensor.dmatrix
        :param input: a symbolic tensor of shape (n_examples, n_in)

        :type n_in: int
        :param n_in: dimensionality of input

        :type n_out: int
        :param n_out: number of hidden units

        :type activation: theano.Op or function
        :param activation: Non linearity to be applied in the hidden
                           layer
        """
	self.input = input
        # end-snippet-1

        # `W` is initialized with `W_values` which is uniformely sampled
        # from sqrt(-6./(n_in+n_hidden)) and sqrt(6./(n_in+n_hidden))
        # for tanh activation function
        # the output of uniform if converted using asarray to dtype
        # theano.config.floatX so that the code is runable on GPU
        # Note : optimal initialization of weights is dependent on the
        #        activation function used (among other things).
        #        For example, results presented in [Xavier10] suggest that you
        #        should use 4 times larger initial weights for sigmoid
        #        compared to tanh
        #        We have no info for other function, so we use the same as
        #        tanh.
        if str(parameters)==str("None"):
            if W is None:
                W_values = numpy.asarray(
                    rng.uniform(
                        low=-numpy.sqrt(6. / (n_in + n_out)),
                        high=numpy.sqrt(6. / (n_in + n_out)),
                        size=(n_in, n_out)
                    ),
                    dtype=theano.config.floatX
                )
                if activation == theano.tensor.nnet.sigmoid:
                    W_values *= 4

            self.W = theano.shared(value=W_values, name='W', borrow=True)

            if b is None:
                b_values = numpy.zeros((n_out,), dtype=theano.config.floatX)
                self.b = theano.shared(value=b_values, name='b', borrow=True)
        else:
            self.W, self.b = cPickle.load(open(str(parameters)+'.pkl')).params
            print("Parameters from "+str(parameters)+" extracted.")

        lin_output = T.dot(input, self.W) + self.b
        self.output = (
            lin_output if activation is None
            else activation(lin_output)
        )
        # parameters of the model
        self.params = [self.W, self.b]



def Size(image_size,filter_size,pool_size):
    return (int((image_size-filter_size+1)/pool_size))


def CorrectSizeCheck(operation_string, filter_size, pool_size, input_dim_x, input_dim_y, input_dim_d):
    operations = numpy.array(operation_string.split(",")).astype('int')
    dim_x = input_dim_x
    dim_y = input_dim_y
    dim_d = input_dim_d

    for i,op in enumerate(operations):
        if op==1:
            dim_x=int((dim_x-filter_size[i][1]+1)/pool_size[i][1])
            dim_y=int((dim_y-filter_size[i][2]+1)/pool_size[i][2])
            dim_d=int((dim_d-filter_size[i][0]+1)/pool_size[i][0])
        if op==0 or op==4 or op==40:
            continue
    print("Final data dimensions are: x: "+str(dim_x)+", y: "+str(dim_y)+", d: "+str(dim_d))
    if dim_x < 1 or dim_y < 1 or dim_d < 1:
        return False
    else:
	return True


def make_model(operations):
    classifier = [0]*len(operations)
    layer = [0]*len(operations)
    for i,op in enumerate(operations):
        if int(op)==4:
            classifier[i] = cPickle.load(open('layer_'+str(i)+'.pkl'))
            layer[i] = theano.function(inputs=[classifier[i].input], outputs=[classifier[i].y_pred])
            print('Final logistic regression layer made.')
            break
        if int(op)==40:
            classifier[i] = cPickle.load(open('layer_'+str(i)+'.pkl'))
            layer[i] = theano.function(inputs=[classifier[i].input], outputs=[classifier[i].y_pred])
       	    print('Final linear regression layer made.')
            break
        else:
            classifier[i] = cPickle.load(open('layer_'+str(i)+'.pkl'))
            layer[i] = theano.function(inputs=[classifier[i].input], outputs=[classifier[i].output])
       	    print('Layer '+str(i)+' made.')            
    return layer

def make_predictions(n, test_set_x, batchsize, operations, nkerns, in_dim_d, filter_size, in_dim_x, in_dim_y, pool_size, model):

    layer=[0]*len(operations)
    predicted_values=numpy.zeros(n*batchsize)

    for i in xrange(n):
        nkern=1
        input = numpy.array(test_set_x[i*batchsize:(i+1)*batchsize]).reshape(batchsize, in_dim_d, nkern, in_dim_x, in_dim_y)
        dim_d=in_dim_d
        dim_x=in_dim_x
        dim_y=in_dim_y
        for j,op in enumerate(operations):
       	    if int(op)==1:
                input=numpy.array(input).reshape(batchsize, dim_d, nkern, dim_x, dim_y)
                layer[j]=model[j](input)
                input=numpy.array(layer[j])
                dim_x=Size(dim_x,filter_size[j][1],pool_size[j][1])
                dim_y=Size(dim_y,filter_size[j][2],pool_size[j][2])
                dim_d=Size(dim_d,filter_size[j][0],pool_size[j][0])
                nkern=nkerns[j]

            if int(op)==0:
                input=numpy.array(input).reshape(batchsize, -1)
                layer[j]=model[j](input)
                input=numpy.array(layer[j])
                nkern=nkerns[j]

            if int(op)==4 or int(op)==40:
                input=numpy.array(input).reshape(batchsize, nkern)
                layer[j]=model[j](input)
                predicted_values[i*batchsize:(i+1)*batchsize]=numpy.array(layer[j])

    return predicted_values


def My_Prediction(operations, nkerns, datafile, batchsize, in_dim_d, filter_size, in_dim_x, in_dim_y, pool_size):

    model=make_model(operations)

    datasets = load_data(datafile)

    test_set_x, test_set_y = datasets[2]
    test_set_x = test_set_x.get_value()
    test_set_y = test_set_y.eval()

    n = int(len(test_set_y)/batchsize)

    y = numpy.squeeze(numpy.array(test_set_y).astype('int'))
    predicted_values = numpy.array(make_predictions(n, test_set_x, batchsize, operations, nkerns, in_dim_d, filter_size, in_dim_x, in_dim_y, pool_size, model)).astype('int')
    print("Y: ",y[-10:])
    print("Predicted values:",predicted_values[-10:])
    if int(nkerns[-1])==5:
        emat = numpy.zeros((5,5))
        tots = numpy.zeros(5)
        for i in xrange(5):
            tots[i] = numpy.sum(y==i)
            for j in xrange(5):
                emat[i,j] = numpy.sum((y==i)*(predicted_values==j))/float(tots[i])
        print('identified as')
        print('            noise - sine-gaus - ringdown - white - bbh')  
        print('noise    {}'.format(emat[0,:]))
        print('sin-gaus {}'.format(emat[1,:]))
        print('ringdown {}'.format(emat[2,:]))
        print('white    {}'.format(emat[3,:])) 
        print('bbh  {}'.format(emat[4,:]))
        print(tots)
    elif int(nkerns[-1])==2:
        emat = numpy.zeros((2,2))
        tots = numpy.zeros(2)
        for i in xrange(2):
            tots[i] = numpy.sum(y==i)
            for j in xrange(2):
                emat[i,j] = numpy.sum((y==i)*(predicted_values==j))/float(tots[i])
        print('identified as')
        print('            bbh - noise')	 
        print('bbh    {}'.format(emat[0,:]))
        print('noise  {}'.format(emat[1,:]))
        print(tots)   # bbh = 0, noise = 1.

    numpy.set_printoptions(threshold='nan')
    results=emat
    print('Creating results.data...')
    numpy.savetxt('results.data',results,fmt='%f')


def My_MSEPrediction(operations, nkerns, datafile, batchsize, in_dim_d, filter_size, in_dim_x, in_dim_y, pool_size):

    print('Making model...')
    model=make_model(operations)

    datasets = load_data(datafile)
    
    print('Loading datasets...')
    test_set_x, test_set_y = datasets[2]
    test_set_x = test_set_x.get_value()
    test_set_y = test_set_y.eval()

    n = int(len(test_set_y)/batchsize)

    y = numpy.squeeze(numpy.array(test_set_y).astype('int'))
    print('Predicting...')
    predicted_values = numpy.array(make_predictions(n, test_set_x, batchsize, operations, nkerns, in_dim_d, filter_size, in_dim_x, in_dim_y, pool_size, model)).astype('float')
    numpy.set_printoptions(threshold='nan')
    print('Creating results.data...')
    results=numpy.column_stack((predicted_values,y))
    numpy.savetxt('results.data',results,fmt='%f')

def load_data(dataset,step=1):
   """ Load the data set

   :param dataset: the path to the data set (string)
   :return: tuple of theano data set
   """
   f = open(dataset, 'rb')
   train_set, valid_set, test_set = cPickle.load(f)
   f.close()

   def shared_dataset(data_xy, borrow=True):
       """ Function that loads the dataset into shared variables

       Shared for GPU acceleration.
       :param data_xy: data and labels in numpy.array
       :param borrow:
       """
       data_x, data_y = data_xy
       shared_x = theano.shared(numpy.asarray(data_x[::step],
                                              dtype=theano.config.floatX),
                                borrow=borrow)
       shared_y = theano.shared(numpy.asarray(data_y[::step],
                                              dtype=theano.config.floatX),
                                borrow=borrow)
       return shared_x, tensor.cast(shared_y, 'int32')

   test_set_x, test_set_y = shared_dataset(test_set)
   valid_set_x, valid_set_y = shared_dataset(valid_set)
   train_set_x, train_set_y = shared_dataset(train_set)

   rval = [(train_set_x, train_set_y),
           (valid_set_x, valid_set_y),
           (test_set_x, test_set_y)]

   return rval

def load_data_zipped(dataset):
    ''' Loads the dataset

    :type dataset: string
    :param dataset: the path to the dataset (here MNIST)
    '''

    #############
    # LOAD DATA #
    #############

    # Download the MNIST dataset if it is not present
    data_dir, data_file = os.path.split(dataset)
    if data_dir == "" and not os.path.isfile(dataset):
        # Check if dataset is in the data directory.
        new_path = os.path.join(
            os.path.split(__file__)[0],
            ".",
            "data",
            dataset
        )
        if os.path.isfile(new_path) or data_file == 'mnist.pkl.gz':
            dataset = new_path

    if (not os.path.isfile(dataset)) and data_file == 'mnist.pkl.gz':
        from six.moves import urllib
        origin = (
            'http://www.iro.umontreal.ca/~lisa/deep/data/mnist/mnist.pkl.gz'
        )
        print('Downloading data from %s' % origin)
        urllib.request.urlretrieve(origin, dataset)

    print('... loading data')

    # Load the dataset
    with gzip.open(dataset, 'rb') as f:
        try:
            train_set, valid_set, test_set = pickle.load(f, encoding='latin1')
        except:
            train_set, valid_set, test_set = pickle.load(f)
    # train_set, valid_set, test_set format: tuple(input, target)
    # input is a numpy.ndarray of 2 dimensions (a matrix)
    # where each row corresponds to an example. target is a
    # numpy.ndarray of 1 dimension (vector) that has the same length as
    # the number of rows in the input. It should give the target
    # to the example with the same index in the input.

    def shared_dataset(data_xy, borrow=True):
        """ Function that loads the dataset into shared variables

        The reason we store our dataset in shared variables is to allow
        Theano to copy it into the GPU memory (when code is run on GPU).
        Since copying data into the GPU is slow, copying a minibatch everytime
        is needed (the default behaviour if the data is not in a shared
        variable) would lead to a large decrease in performance.
        """
        data_x, data_y = data_xy
        shared_x = theano.shared(numpy.asarray(data_x,
                                               dtype=theano.config.floatX),
                                 borrow=borrow)
        shared_y = theano.shared(numpy.asarray(data_y,
                                               dtype=theano.config.floatX),
                                 borrow=borrow)
        # When storing data on the GPU it has to be stored as floats
        # therefore we will store the labels as ``floatX`` as well
        # (``shared_y`` does exactly that). But during our computations
        # we need them as ints (we use labels as index, and if they are
        # floats it doesn't make sense) therefore instead of returning
        # ``shared_y`` we will have to cast it to int. This little hack
        # lets ous get around this issue
        return shared_x, T.cast(shared_y, 'int32')

    test_set_x, test_set_y = shared_dataset(test_set)
    valid_set_x, valid_set_y = shared_dataset(valid_set)
    train_set_x, train_set_y = shared_dataset(train_set)

    rval = [(train_set_x, train_set_y), (valid_set_x, valid_set_y),
            (test_set_x, test_set_y)]
    return rval




class LeNetConvPool3DLayer(object):
    """Pool Layer of a convolutional network """

    def __init__(self, rng, input, filter_shape, image_shape, poolsize=(2, 2, 2), activation=T.tanh, reduce_shape=False, parameters=str("None")):
        """
        Allocate a LeNetConvPoolLayer with shared variable internal parameters.

        :type rng: numpy.random.RandomState
        :param rng: a random number generator used to initialize weights

        :type input: theano.tensor.dtensor4
        :param input: symbolic image tensor, of shape image_shape

        :type filter_shape: tuple or list of length 4
        :param filter_shape: (number of filters, num input feature maps,
                              filter height, filter width)

        :type image_shape: tuple or list of length 4
        :param image_shape: (batch size, num input feature maps,
                             image height, image width)

        :type poolsize: tuple or list of length 2
        :param poolsize: the downsampling (pooling) factor (#rows, #cols)
        """

        assert image_shape[2] == filter_shape[2]
        self.input = input

        # there are "num input feature maps * filter height * filter width"
        # inputs to each hidden unit
        fan_in = numpy.prod(filter_shape[1:])
        # each unit in the lower layer receives a gradient from:
        # "num output feature maps * filter height * filter width" /
        #   pooling size
        fan_out = (filter_shape[0] * numpy.prod(filter_shape[3:] * filter_shape[1]) //
                   numpy.prod(poolsize))
        # initialize weights with random weights
        
        if str(parameters) == str("None"):
            W_bound = numpy.sqrt(6. / (fan_in + fan_out))
            self.W = theano.shared(
                numpy.asarray(
                    rng.uniform(low=-W_bound, high=W_bound, size=filter_shape),
                    dtype=theano.config.floatX
                ),
                borrow=True
            )

            # the bias is a 1D tensor -- one bias per output feature map
            b_values = numpy.zeros((filter_shape[0],), dtype=theano.config.floatX)
            self.b = theano.shared(value=b_values, borrow=True)

        else:
            self.W, self.b = cPickle.load(open(str(parameters)+'.pkl')).params
            print("Parameters from "+str(parameters)+" extracted.")


        # convolve input feature maps with filters
        conv_out = conv3d(
            input,
            self.W,
            signals_shape=image_shape,
            filters_shape=filter_shape
        )

        # pool each feature map individually, using maxpooling
        pooled_out = my_pool_3d(conv_out, poolsize, True)

        # add the bias term. Since the bias is a vector (1D array), we first
        # reshape it to a tensor of shape (1, n_filters, 1, 1). Each bias will
        # thus be broadcasted across mini-batches and feature map
        # width & height
        if reduce_shape:
            print("Reducing output dimensions...")
            if ((image_shape[1]-filter_shape[1]+1)/poolsize[0])<1:
                lin_output = T.tanh(pooled_out + self.b.dimshuffle('x', 'x', 0, 'x', 'x'))
            else:
                lin_output = T.tanh(pooled_out + self.b.dimshuffle('x', 'x', 0, 'x', 'x')).dimshuffle(1,0,2,3,4)[0]
        else:
            lin_output = T.tanh(pooled_out + self.b.dimshuffle('x', 'x', 0, 'x', 'x'))
        # store parameters of this layer
        self.output = activation(lin_output)
        self.params = [self.W, self.b]

        # keep track of model input
        self.input = input


def my_pool_3d(pool_in, dw=(2, 2, 2), ignore_border_=True):

    pool_out_1 = pool.pool_2d(input=pool_in.dimshuffle(0,2,4,3,1), ds=(dw[0], dw[1]), ignore_border=ignore_border_)
    pool_out = pool.pool_2d(input=pool_out_1.dimshuffle(0, 1, 4, 3, 2), ds=(1, dw[2]), ignore_border=ignore_border_)

    return pool_out.dimshuffle(0, 2, 1, 3, 4)


